#include "limit.h"

// The parameters that shouldn't be hard-wired in, but are

const int max_iterations = 100;
const double max_x = 100;
const double min_x = -100;
const double max_y = 100;
const double min_y = -100;
const double side_buffer = 1;
const double resolution = 1E-1;
const Complex base_point = {0.5, -0.2886751345948};

bool point_within_box(Complex z){
  double r, i;
  r = z.real;
  i = z.imag;
  return (r < max_x + side_buffer) && (r > min_x - side_buffer) &&
    (i < max_y + side_buffer) && (i > min_y - side_buffer);
}

bool point_within_show_box(Complex z){
  double r, i;
  r = z.real;
  i = z.imag;
  return (r < max_x) && (r > min_x) &&
    (i < max_y) && (i > min_y);
}


bool points_are_close(PointRecord& p1, PointRecord& p2){
  return (fabs(p1.z.real - p2.z.real) < resolution) &&
    (fabs(p1.z.imag - p2.z.imag) < resolution);
}
    

int main(int argc, char* argv[]){
  MoebiusTransformation U, V, P, Uinv, Vinv;
  MoebiusTransformation generators[4];
  int dir_entered[4];
  Complex z;
  list<PointRecord> the_points;
  list<PointRecord> new_points;
  int i, j, list_len;
  list<PointRecord>::iterator p;
  list<PointRecord>::iterator q;
  
  fiber_generators(generators, dir_entered);
  
  // setup initial 4 points

  for (i=0; i < 4; i++){
    z = apply_mt_to_bdry(&generators[i], base_point);
    PointRecord point = PointRecord(&generators[i], z, i);
    the_points.push_back(point);
  }
  the_points.sort();
  list_len = the_points.size();

  for (i=0; i < max_iterations; i++){
    for(p = the_points.begin(); p != the_points.end(); ++p){
      if (p->is_new){
	p->is_new = false;
	for (j=0; j < 4; j++){
	  if( j != p->dir_entered){
	    Moebius_product(&p->group_elt, &generators[j], &P);
	    z = apply_mt_to_bdry(&P, base_point);
	    if ( point_within_box(z) ){
	      PointRecord new_point = PointRecord(&P, z, dir_entered[j]);
	      new_points.push_back(new_point);
	    }
	  }
	}
      }
    }
    new_points.sort();
    the_points.merge(new_points);

    // prune (using that the data is sorted)
    
    for (p = the_points.begin(); p != the_points.end(); ++p){
      q = p;
      q++;
      while( q != the_points.end() ){
	if (points_are_close(*p, *q)){
	  q = the_points.erase(q);
	}
	else
	  q++;
	if (fabs(p->z.real - q->z.real) > resolution)
	  break;
      }
    }
    
    if (the_points.size()  == list_len)
      break;
    
    list_len = the_points.size();
    cerr << i << "  " << the_points.size() << "\n";

  }

  cout.precision(15);
  for(p = the_points.begin(); p != the_points.end(); ++p){
    if (point_within_show_box(p->z))
      cout << p->z.real << "\t" << p->z.imag << "\n";
  }


}
