#include "limit.h"

void print_complex(Complex z){
  cout << z.real <<  " + " << z.imag << " I";
}

void print_mt(MoebiusTransformation* M){
  cout << "MoebiusTransformation:  ";
  if (M->parity == orientation_preserving)
    cout << "orient. preserving\n";
  else
    cout << "orient. reverving\n";
  
  cout << "   ";
  print_complex(M->matrix[0][0]);
  cout  << "\t";
  print_complex(M->matrix[0][1]);
  cout  << "\n    ";
  print_complex(M->matrix[1][0]);
  cout  << "\t";
  print_complex(M->matrix[1][1]);
  cout  << "\n";
}

// follows convention of Alperin-Dicks-Porti

void gens_of_gieseking_manifold(MoebiusTransformation* U,
				MoebiusTransformation* V){
  Complex omega, omega_sq;
  Complex alpha = {-1,2};
  MoebiusTransformation P, Pinv, T;

  /*
  P.parity = orientation_preserving;
  P.matrix[0][0] = complex_negate(complex_mult(alpha, alpha));
  P.matrix[0][1] = One;
  P.matrix[1][0] = alpha;
  P.matrix[1][1] = Zero;
  Moebius_invert(&P, &Pinv);
  */

  Complex b = {3.61476, 0};
  Complex c = {0, 2.07665};
  Complex d = {2.15555, 0};
  P.parity = orientation_preserving;
  P.matrix[0][0] = Zero;
  P.matrix[0][1] = b;
  P.matrix[1][0] = c;
  P.matrix[1][1] = d;
  Moebius_invert(&P, &Pinv);

  //Initialize:
  
  omega = complex_exp(complex_real_mult(1/3.0, TwoPiI));
  omega_sq = complex_exp(complex_real_mult(2/3.0, TwoPiI));

  U->parity = orientation_reversing;
  U->matrix[0][0] = Zero;
  U->matrix[0][1] = omega_sq;
  U->matrix[1][0] = complex_negate(omega);
  U->matrix[1][1] = omega_sq;
  
  V->parity = orientation_reversing;
  V->matrix[0][0] = omega;
  V->matrix[0][1] = omega_sq;
  V->matrix[1][0] = Zero;
  V->matrix[1][1] = omega_sq;

  // Change coor

  Moebius_product(&P, U, &T);
  Moebius_product(&T, &Pinv , U);

  Moebius_product(&P, V, &T);
  Moebius_product(&T, &Pinv , V);
  
  
}

void old_gens_of_gieseking_manifold(MoebiusTransformation* U,
				MoebiusTransformation* V){
  Complex omega, omega_sq;

  //Initialize:
  
  omega = complex_exp(complex_real_mult(1/3.0, TwoPiI));
  omega_sq = complex_exp(complex_real_mult(2/3.0, TwoPiI));

  U->parity = orientation_reversing;
  U->matrix[0][0] = Zero;
  U->matrix[0][1] = omega_sq;
  U->matrix[1][0] = complex_negate(omega);
  U->matrix[1][1] = omega_sq;
  
  V->parity = orientation_reversing;
  V->matrix[0][0] = omega;
  V->matrix[0][1] = omega_sq;
  V->matrix[1][0] = Zero;
  V->matrix[1][1] = omega_sq;
}

void fiber_generators(MoebiusTransformation* generators, int* dir_entered){

  MoebiusTransformation U, V, P, Uinv, Vinv;
  
  // setup movements

  gens_of_gieseking_manifold(&U,&V);
  Moebius_invert(&U, &Uinv);
  Moebius_invert(&V, &Vinv);

  Moebius_product(&U, &V, &P);
  Moebius_copy(&generators[0], &P);
  dir_entered[0] = 2;

  Moebius_product(&V, &U, &P);
  Moebius_copy(&generators[1], &P);
  dir_entered[1] = 3;

  Moebius_product(&Vinv, &Uinv, &P);
  Moebius_copy(&generators[2], &P);
  dir_entered[2] = 0;

  Moebius_product(&Uinv, &Vinv, &P);
  Moebius_copy(&generators[3], &P);
  dir_entered[3] = 1;
}

void fiber_section_generators(MoebiusTransformation* generators, int* dir_entered){

  MoebiusTransformation U, V, P, Uinv, Vinv;
  
  // setup movements

  gens_of_gieseking_manifold(&U,&V);
  Moebius_invert(&U, &Uinv);
  Moebius_invert(&V, &Vinv);

  Moebius_product(&U, &V, &P);
  Moebius_copy(&generators[0], &P);
  dir_entered[0] = 2;

  Moebius_product(&Vinv, &Vinv, &P);
  Moebius_product(&P, &Uinv, &P);
  Moebius_copy(&generators[1], &P);
  dir_entered[1] = 0;

  Moebius_product(&Vinv, &Uinv, &P);
  Moebius_copy(&generators[2], &P);
  dir_entered[2] = 0;

  Moebius_product(&U, &U, &P);
  Moebius_product(&P, &V, &P);
  Moebius_copy(&generators[3], &P);
  dir_entered[3] = 2;
}

void other_section_generators(MoebiusTransformation* generators, int* dir_entered){

  MoebiusTransformation U, V, P, Uinv, Vinv;
  
  // setup movements

  gens_of_gieseking_manifold(&U,&V);
  Moebius_invert(&U, &Uinv);
  Moebius_invert(&V, &Vinv);

  Moebius_product(&Uinv, &Vinv, &P);
  Moebius_copy(&generators[0], &P);
  dir_entered[0] = 1;

  Moebius_product(&U, &Vinv, &P);
  Moebius_copy(&generators[1], &P);
  dir_entered[1] = 2;

  Moebius_product(&V, &Uinv, &P);
  Moebius_copy(&generators[2], &P);
  dir_entered[2] = 1;

  Moebius_product(&V, &V, &P);
  Moebius_product(&P, &Uinv, &P);
  Moebius_copy(&generators[3], &P);
  dir_entered[3] = 1;
}

Complex apply_mt_to_bdry(MoebiusTransformation* T, Complex z){
  Complex my_z, top, bot;
  if (T->parity == orientation_reversing)
    my_z = complex_conjugate(z);
  else
    my_z = z;
  top = complex_plus(complex_mult(T->matrix[0][0], my_z), T->matrix[0][1]);
  bot = complex_plus(complex_mult(T->matrix[1][0], my_z), T->matrix[1][1]);
  return complex_div(top, bot);
}
