/*
 *	Census.c
 *     
 *     Modified from Census.cp by Nathan Dunfield
 *     (nathand@math.uchicago.edu) and others to work under unix.
 *     Assumes the existence of the environment variable:
 *      
 *     SNAPPEA_TRS_DIRECTORY 
*/

#include "Census.h"
#include "CensusResourceIDs.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  /* needed for SEEK_END, SEEK_SET which sometimes may 
		      * be in stdio.h
		      */

extern void *my_malloc(size_t bytes);
extern void my_free(void *ptr);

Triangulation *GetCensusManifold(
	short	aCensus,
	short	anIndex,
	char*   census_directory)
{
	int						theNumCensusManifolds,
							theResourceID;
	FILE *					dataFile;
	char					*theData;
	TersestTriangulation	*theTersest;
	Triangulation			*theTriangulation = NULL;
	char					theName[10];
	const char		*dirName;
	char	fileName[500];
	long int 	fileSize;
	
	switch (aCensus)
	{
		case 5:

			theNumCensusManifolds	= 415;
			theResourceID			= five_tet_tersest_ID;
			sprintf(theName, "m%.3d", anIndex);
			break;

		case 6:

			theNumCensusManifolds	= 962;
			theResourceID			= six_tet_tersest_ID;
			sprintf(theName, "s%.3d", anIndex);
			break;

		case 8:		/* really the nonorientable 6-tet census */

			theNumCensusManifolds	= 259;
			theResourceID			= six_tet_nonor_tersest_ID;
			sprintf(theName, "x%.3d", anIndex);
			break;

		case 7:

			theNumCensusManifolds	= 3552;
			theResourceID			= seven_tet_tersest_ID;
			sprintf(theName, "v%.4d", anIndex);
			break;

		case 9:		/* really the nonorientable 7-tet census */

			theNumCensusManifolds	= 887;
			theResourceID			= seven_tet_nonor_tersest_ID;
			sprintf(theName, "y%.3d", anIndex);
			break;

		default:
			uFatalError("GetCensusManifold", "Census");
	}

	if (anIndex < 0 || anIndex >= theNumCensusManifolds)
		return NULL;

	dirName = census_directory;
	if (dirName == NULL)
		return NULL;
	
	strcpy(fileName, dirName);
	strcat(fileName, "/");
	{ char resIDStr[100];
	sprintf(resIDStr, "%d", (int)theResourceID);
	strcat(fileName, resIDStr);
	}
	
	dataFile = fopen(fileName, "rb");
	if (dataFile != NULL){
	  (int) fseek(dataFile, 0, SEEK_END);
	  fileSize = ftell(dataFile);
	  if (fileSize > 0){
	  theData = (char *) my_malloc(fileSize);
	  if (theData !=  NULL){
	  (int) fseek(dataFile, 0, SEEK_SET);
	  if (
	      0 != fread(theData, 1, fileSize, dataFile) ) {
	    theTersest = (TersestTriangulation *)
	      ((char*)theData + anIndex * sizeof(TersestTriangulation));
	    rehydrate_census_manifold(*theTersest, aCensus, anIndex, &theTriangulation);
	    set_triangulation_name(theTriangulation, theName);
	  }
	}
	}
	  fclose(dataFile);
	}

	my_free(theData);

	return theTriangulation;
}


