// Typemaps used for the Python SnapPea interface.

%{
#include "kernel.h"
%}

// FuncResult is no longer returned -- It just raises an exception if something fails.

typedef int FuncResult;
enum
{
	func_OK = 0,
	func_cancelled,
	func_failed,
	func_bad_input
};

%typemap(python, out) FuncResult {
	if($source == func_cancelled){
		PyErr_SetString(PyExc_KeyboardInterrupt, "User interuped SnapPea computation");
	       	return NULL;
	}
	if($source == func_failed){
		PyErr_SetString(PyExc_RuntimeError, "Internal SnapPea computation failed");
	       	return NULL;
	}
	if($source == func_bad_input){
		PyErr_SetString(PyExc_ValueError, "Bad input to internal SnapPea function");
	       	return NULL;
	}
        Py_INCREF(Py_None);
	$target = Py_None;
}

// Typemap for Triangulation* so that you can't return NULL

//%typemap(python, out) Triangulation* {
//	if ($source)
//		return N_Make_SWIG_Ptr_Object((char*) $source, "_Triangulation_p");
//	else{
//		PyErr_SetString(PyExc_RuntimeError, "NULL Triangulation returned");
//		return NULL;
//	}
//}

// Typemaps for SnapPea Complex type to Python Complex type

%typemap(python, out) Complex {
	$target = PyComplex_FromDoubles($source->real, $source->imag);
	free($source);
}

%typemap(python, in) Complex* (Complex temp_complex){
	if(PyComplex_Check($source) ){
		temp_complex.real = PyComplex_RealAsDouble($source);
		temp_complex.imag = PyComplex_ImagAsDouble($source);
		$target = &temp_complex;
	}
	else{
		PyErr_SetString(PyExc_TypeError,"Need a complex number");
	       	return NULL;
	}
}

// Below are modeled on typemaps.i. 

%typemap(python, ignore) Complex *OUTPUT(Complex temp){
	$target = &temp;
}

%typemap(python, argout) Complex *OUTPUT {
	PyObject *o;
	o = PyComplex_FromDoubles($source->real, $source->imag);
	$target = t_output_helper($target, o);
}


// The next set of type maps deal with functions that return new structures 
// via ** arguements.  If it weren't for the bug with $basename macro, 
// I could do this all with a single typemap.  (*sigh*).  

%typemap(python, ignore) IsometryList** OUTPUT (IsometryList* temp ){
	temp = NULL;
	$target = &temp;
}
%typemap(python, argout) IsometryList **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_IsometryList_p" ));
}

%typemap(python, ignore) NormalSurfaceList** OUTPUT (NormalSurfaceList* temp ){
	temp = NULL;
	$target = &temp;
}
%typemap(python, argout) NormalSurfaceList **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_NormalSurfaceList_p" ));
}

%typemap(python, ignore) LRFactorization** OUTPUT (LRFactorization* temp ){
	temp = NULL;
	$target = &temp;
}
%typemap(python, argout) LRFactorization **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_LRFactorization_p" ));
}

%typemap(python, ignore) AbelianGroup** OUTPUT (AbelianGroup* temp ){
	temp = NULL;
	$target = &temp;
}
%typemap(python, argout) AbelianGroup **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_AbelianGroup_p" ));
}


%typemap(python, ignore) SymmetryGroup** OUTPUT (SymmetryGroup* temp ){
	temp = NULL;
	$target = &temp;
}
%typemap(python, argout) SymmetryGroup **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_SymmetryGroup_p" ));
}

%typemap(python, ignore) Triangulation** OUTPUT (Triangulation* temp ){
	temp = NULL;
	$target = &temp;
}

%typemap(python, argout) Triangulation **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_Triangulation_p" ));
}


%typemap(python, ignore) DualOneSkeletonCurvePtr** OUTPUT (DualOneSkeletonCurvePtr* temp ){
	temp = NULL;
	$target = &temp;
}

%typemap(python, argout) DualOneSkeletonCurvePtr **OUTPUT{
	$target = t_output_helper($target, N_Make_SWIG_Ptr_Object((char* ) *$source, "_DualOneSkeletonCurvePtr_p" ));
}


