from SnapPea import *
from paths import *
import os, re

# "interators" for the various censuses.

class oriented_six_census:
    def __init__(self, *d):
        if len(d) == 0:
            self.start = 0
            self.end = 962
        elif len(d) == 1:
            self.start = 0
            self.end = d[0]
        elif len(d) == 2:
            self.start = d[0]
            self.end = d[1]
        else:
            raiseTypeError; "too many arguments; expect 2 or fewer"
        
    def __getitem__(self, i):
        curr_index = i + self.start
        if not 0 <= curr_index<=  961: raise IndexError
        if not curr_index < self.end: raise IndexError
        return GetCensusManifold(6, curr_index)

class oriented_seven_census:
    def __init__(self, *d):
        if len(d) == 0:
            self.start = 0
            self.end = 3552
        elif len(d) == 1:
            self.start = 0
            self.end = d[0]
        elif len(d) == 2:
            self.start = d[0]
            self.end = d[1]
        else:
            raiseTypeError; "too many arguments; expect 2 or fewer"
        
    def __getitem__(self, i):
        curr_index = i + self.start
        if not 0 <= curr_index<=  3551: raise IndexError
        if not curr_index < self.end: raise IndexError
        return GetCensusManifold(7, curr_index)

class five_census:
    def __init__(self, *d):
        if len(d) == 0:
            self.start = 0
            self.end = 415
        elif len(d) == 1:
            self.start = 0
            self.end = d[0]
        elif len(d) == 2:
            self.start = d[0]
            self.end = d[1]
        else:
            raiseTypeError; "too many arguments; expect 2 or fewer"
        
    def __getitem__(self, i):
        curr_index = i + self.start
        if not 0 <= curr_index<=  414: raise IndexError
        if not curr_index < self.end: raise IndexError
        return GetCensusManifold(5, curr_index)

class oriented_five_census:
    orientable_fives = [3, 4, 6, 7, 9, 10, 11, 15, 16, 17, 19, 22, 23, 26, 27, 29, 30, 32,
                        33, 34, 35, 36, 37, 38, 39, 40, 43, 44, 45, 46, 47, 49, 52, 53, 54,
                        55, 58, 59, 60, 61, 62, 64, 66, 67, 69, 70, 71, 72, 73, 74, 76, 77,
                        78, 79, 80, 81, 82, 83, 84, 85, 87, 89, 90, 93, 94, 95, 96, 98, 99,
                        100, 102, 103, 104, 105, 106, 108, 110, 111, 112, 113, 115, 116, 117,
                        118, 119, 120, 121, 122, 123, 125, 129, 130, 135, 136, 137, 139, 140,
                        141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 154, 155, 157,
                        159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172,
                        173, 175, 178, 179, 180, 181, 182, 183, 184, 185, 186, 188, 189, 190,
                        191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 206,
                        207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220,
                        221, 222, 223, 224, 225, 227, 228, 229, 230, 231, 232, 233, 234, 235,
                        236, 238, 239, 240, 241, 242, 243, 244, 246, 247, 248, 249, 250, 251,
                        252, 253, 255, 256, 257, 258, 259, 260, 261, 262, 263, 266, 267, 269,
                        270, 271, 272, 273, 275, 276, 277, 278, 279, 280, 281, 282, 284, 285,
                        286, 287, 288, 289, 290, 291, 292, 293, 294, 295, 296, 297, 299, 300,
                        302, 303, 304, 305, 306, 307, 310, 312, 316, 318, 319, 320, 322, 326,
                        328, 329, 335, 336, 337, 338, 339, 340, 342, 343, 345, 346, 349, 350,
                        351, 352, 356, 357, 358, 359, 360, 361, 363, 364, 366, 367, 368, 369,
                        370, 371, 372, 373, 374, 376, 378, 385, 388, 389, 390, 391, 392, 393,
                        395, 397, 400, 401, 402, 403, 410, 412]

    def __init__(self, *d):
        if len(d) == 0:
            self.start = 0
            self.end = 301
        elif len(d) == 1:
            self.start = 0
            self.end = oriented_five_census.orientable_fives.index(d[0])
        elif len(d) == 2:
            self.start = oriented_five_census.orientable_fives.index(d[0])
            self.end = oriented_five_census.orientable_fives.index(d[1]) 
        else:
            raiseTypeError; "too many arguments; expect 2 or fewer"
        
    def __getitem__(self, i):
        curr_index = i + self.start
        if not 0 <= curr_index<=  301: raise IndexError
        if not curr_index < self.end: raise IndexError
        return GetCensusManifold(5, oriented_five_census.orientable_fives[curr_index])


# Iterators for knots and links


class link_complements:
    
    # num_links[component][crossings] is the number of links with
    # specified number of components and crossings.
    num_links= [ [], [0, 0, 0, 1, 1, 2, 3, 7, 21, 49, 166, 552],
                 [0, 0, 0, 0, 1, 1, 3, 8, 16, 61, 183],
                 [0, 0, 0, 0, 0, 0, 3, 1, 10, 21, 74],
                 [0, 0, 0, 0, 0, 0, 0, 0, 3, 1, 24],
                 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3]];

    def __init__(self, components, crossings=12):
        if not (1 <= components < len(self.num_links) ):
            raise IndexError, "No data on links with %s many components." % components
        self.components = components
        self.crossings = crossings
        max = 0
        for n in self.num_links[components][:crossings]:
            max = max + n
        self.max = max

    def __getitem__(self, i):
        if not (0 <= i < self.max):
            raise IndexError, "Requested link complement out of range"
        so_far = 0
        for k in range(self.crossings):
            n =  self.num_links[self.components][k]
            so_far = so_far + n
            if so_far > i:
                j = i - so_far + n + 1
                if self.components > 1:
                    return get_manifold("%d^%d_%d" % (k, self.components, j))
                else:
                    return get_manifold("%d_%d" % (k,  j))
                                   
#--- end link complements ---------------------------------------

class orientable_closed_manifolds:
    def __init__(self, max = 100000):
        f = open(closed_manifold_directory + os.sep + "closed_manifolds")
        data = []
        for line in f.readlines():
            m = re.search("(\[.*\])", line)
            name = eval(m.group(1))[0]
            data.append(name)
            if len(data) >= max:
                break
        self.data = data
        self.max = min((max, len(data)))

    def __getitem__(self, i):
        if not (0 <= i < self.max):
            raise IndexError, "Requested manifold is out of range."
        return get_manifold(self.data[i])
            
                 









