import SnapPeac
from SnapPea import *

# Class to deal with covers

class CoversOfManifold:
    def __init__(self, manifold, degree, perm_subgroup=permutation_subgroup_Sn):
        self.rep_list = find_representations(manifold, degree, perm_subgroup)
        self.length = len(self.rep_list)
        self.base_manifold = manifold
        self.degree = degree

    def __len__(self):
        return self.length
    
    def __getitem__(self, i):
        if i < 0 or i >= self.length:
            raise IndexError, "Cover index out of range"

        cover, type = self.rep_list.get_cover(i, self.base_manifold)
        cover = TriangulationPtr(cover)
        cover.cover_type = ("unknown", "irregular",
                            "regular",  "cyclic")[type]
        cover.degree = self.degree
        cover.index = i
        set_triangulation_name(cover, get_triangulation_name(self.base_manifold) +
                               "_cover_%i_deg_%i_type_%s" %
                               (cover.index, cover.degree, cover.cover_type))
        cover.thisown = 1
        return cover

    def __repr__(self):
        return "CoversOfManifold %s of degree %i" % (self.base_manifold, self.degree)


# MatrixInt22

def MatrixInt22_entries(self):
    val = SnapPeac.MatrixInt22_entries(self.this)
    return ((val[0], val[1]), (val[2], val[3]))

MatrixInt22Ptr.entries = MatrixInt22_entries

def MatrixInt22_repr(self):
    return "MatrixInt22 " + repr(self.entries())

MatrixInt22Ptr.__repr__ = MatrixInt22_repr

# Moebius Transformation

def MoebiusTransformation_entries(self):
    val = SnapPeac.MoebiusTransformation_entries(self.this)
    return ((val[0], val[1]), (val[2], val[3]))

MoebiusTransformationPtr.entries = MoebiusTransformation_entries

def MoebiusTransformation_repr(self):
    if self.parity == orientation_reversing:
        parity = " orient. rev."
    else:
        parity = " orient. pres."
    return "MoebiusTransformation" +  repr(self.entries()) + parity

MoebiusTransformationPtr.__repr__ = MoebiusTransformation_repr

def MoebiusTransformation_length(self):
    return complex_length_mt(self)

MoebiusTransformationPtr.length = MoebiusTransformation_length

def MoebiusTransformation_trace(self):
    val = SnapPeac.MoebiusTransformation_entries(self.this)
    return val[0] + val[3]
    
MoebiusTransformationPtr.trace = MoebiusTransformation_trace

# O31Matrix

def O31Matrix_entries(self):
    val = SnapPeac.O31Matrix_entries(self.this)
    return ((val[0], val[1], val[2], val[3]), (val[4], val[5], val[6], val[7]), (val[8], val[9], val[10], val[11]), (val[12], val[13], val[14], val[15]))

O31MatrixPtr.entries = O31Matrix_entries

def O31Matrix_repr(self):
    return "O31Matrix" + repr(self.entries())

O31MatrixPtr.__repr__ = O31Matrix_repr

def O31Matrix_length(self):
    return complex_length_o31(self)

O31MatrixPtr.length = O31Matrix_length


# GL4RMatrix

def GL4RMatrix_entries(self):
    val = SnapPeac.GL4RMatrix_entries(self.this)
    return ((val[0], val[1], val[2], val[3]), (val[4], val[5], val[6], val[7]), (val[8], val[9], val[10], val[11]), (val[12], val[13], val[14], val[15]))

GL4RMatrixPtr.entries = GL4RMatrix_entries

def GL4RMatrix_repr(self):
    return "GL4RMatrix" + repr(self.entries())

GL4RMatrixPtr.__repr__ = GL4RMatrix_repr

# O31Vector

def O31Vector_repr(self):
    return "O31Vector" + repr(self.entries())

O31VectorPtr.__repr__ = O31Vector_repr

# AbelianGroup

def AbelianGroup_len(self):
    return self.num_torision_coefficients

AbelianGroupPtr.__len__ = AbelianGroup_len

def AbelianGroup_torsion(self):
    coeffs = self.coefficient_list()
    p = 1
    for c in coeffs:
        if c > 0:
            p = c*p
    return p

AbelianGroupPtr.torsion_order = AbelianGroup_torsion

def AbelianGroup_coefficient_list(self):
    n = self.num_torsion_coefficients
    ptr = self.torsion_coefficients
    val = [0,]*n
    for i in range(0, n):
        val[i] = ptrvalue(ptr, i)
    return val

AbelianGroupPtr.coefficient_list = AbelianGroup_coefficient_list

def AbelianGroup_repr(self):
    val = ""
    coeffs = self.coefficient_list()
    for i in range(0, len(coeffs)):
        if coeffs[i] == 0:
            val = val + "Z"
        else:
            val = val + "Z/%i" % coeffs[i]
        if i != len(coeffs) - 1:
            val = val + " + "

    return val

AbelianGroupPtr.__repr__ = AbelianGroup_repr

# GroupPresentation

def GroupPresentation_word_to_string(word):
    s = ""
    i = 0
    while ptrvalue(word, i) != 0:
        v = ptrvalue(word, i)
        if v > 0:
            s = s + chr(v + 96)
        else:
            s = s + chr(-v + 64)
        i = i + 1
    return s

def GroupPresentation_string_to_word(string):
    word = ptrcreate("int", 0, len(string) + 1)
    for i in range(len(string)):
        o = ord(string[i])
        if o > 96:
            ptrset(word, o - 96, i)
        else:
            ptrset(word, -(o - 64), i)
    return word

def GroupPresentation_num_gens(self):
    return fg_get_num_generators(self)

GroupPresentationPtr.num_gens = GroupPresentation_num_gens

def GroupPresentation_gens(self):
    num = fg_get_num_generators(self)
    return map(lambda x: chr(x + 97), range(num))

GroupPresentationPtr.gens = GroupPresentation_gens


def GroupPresentation_num_rels(self):
    return fg_get_num_relations(self)

GroupPresentationPtr.num_rels = GroupPresentation_num_rels

def GroupPresentation_relation(self, i):
    if not (0 <= i < self.num_rels()):
        raise IndexError, "Group has only %d relations, %d requested." % (self.num_rels(), i)
    word = fg_get_relation(self, i)
    str =  GroupPresentation_word_to_string(word)
    fg_free_relation(word)
    return str

GroupPresentationPtr.relation = GroupPresentation_relation

def GroupPresentation_relations(self):
    rels = []
    for i in range(self.num_rels()):
        rels.append(self.relation(i))
    return rels

GroupPresentationPtr.relations = GroupPresentation_relations

def GroupPresentation_repr(self):
    s = "< "
    gens = self.gens()
    for gen in gens[:-1]:
        s = s + gen + ", "
    s = s + gens[-1] + " | "
    rels = self.relations()
    for rel in rels[:-1]:
        s = s + rel + ", "
    if len(rels) > 0:
        s = s + rels[-1] + " >"
    else:
        s = s + " >"
    return s

GroupPresentationPtr.__repr__ = GroupPresentation_repr

def GroupPresentation_word_to_matrix(self, string):
    word = GroupPresentation_string_to_word(string)
    M = O31Matrix()
    T = MoebiusTransformation()
    fg_word_to_matrix(self, word, M, T)
    ptrfree(word)
    return (M, T)

GroupPresentationPtr.word_to_matrix = GroupPresentation_word_to_matrix

def GroupPresentation_word_to_moebius(self, string):
    return self.word_to_matrix(string)[1]

GroupPresentationPtr.word_to_moebius = GroupPresentation_word_to_moebius

def GroupPresentation_num_cusps(self):
    return fg_get_num_cusps(self)

GroupPresentationPtr.num_cusps = GroupPresentation_num_cusps

def GroupPresentation_meridian(self, i):
    if not (0 <= i < self.num_cusps()):
        raise IndexError, "Cusp index out of range"
    word = fg_get_meridian(self, i)
    str =  GroupPresentation_word_to_string(word)
    fg_free_relation(word)
    return str

GroupPresentationPtr.meridian = GroupPresentation_meridian

def GroupPresentation_longitude(self, i):
    if not (0 <= i < self.num_cusps()):
        raise IndexError, "Cusp index out of range"
    word = fg_get_longitude(self, i)
    str =  GroupPresentation_word_to_string(word)
    fg_free_relation(word)
    return str

GroupPresentationPtr.longitude = GroupPresentation_longitude

def GroupPresentation_integer_fillings(self):
    return fg_get_integer_fillings(self)

GroupPresentationPtr.integer_fillings = GroupPresentation_integer_fillings



# Triangulation

def Triangulation_init(self,this):
    if this == "NULL":
        raise RuntimeError, "Tried to create Triangulation from NULL pointer"
    self.this = this
    self.thisown = 0

TriangulationPtr.__init__ = Triangulation_init

def repr_triangulation(self):
    return get_triangulation_name(self )

TriangulationPtr.__repr__ = repr_triangulation

def covers_of_triangulation(self, degree):
    return CoversOfManifold(self, degree)

TriangulationPtr.covers = covers_of_triangulation

def cyclic_covers_of_triangulation(self, degree):
    return CoversOfManifold(self, degree, permutation_subgroup_Zn)

TriangulationPtr.cyclic_covers = cyclic_covers_of_triangulation

def first_betti_number(self, coeff=0):
    coeffs = homology(self).coefficient_list()
    if coeff != 0:
        coeffs = map(lambda x, b=coeff: x % b, coeffs)
    return coeffs.count(0)

TriangulationPtr.first_betti = first_betti_number

def Triangulation_volume(self):
    return volume(self)[0]

TriangulationPtr.volume = Triangulation_volume

def Triangulation_complete_solution_type(self):
    return get_complete_solution_type(self)

TriangulationPtr.complete_solution_type = Triangulation_complete_solution_type

TriangulationPtr.homology = homology
TriangulationPtr.fundamental_group = fundamental_group

# DualOneSkeletonCurve

#def DualOneSkeletonCurve_info(self):
#    return get_dual_curve_info(self)
#
#DualOneSkeletonCurve.info = DualOneSkeletonCurve_info
#
#def DualOneSkeletonCurve_repr(self):
#    return "unfilled %s   filled %s" % self.info()[:2]
#    
#DualOneSkeletonCurve.__repr__ = DualOneSkeletonCurve_repr

class DualOneSkeletonCurvePtrPtr :
    def __init__(self, this, len, manifold):
        self.this = this
        self.manifold = manifold
        self.len = len
        self.thisown = 1

    def __repr__(self):
        return "<C DualOneSkeletonCurvePtr instance>"


    def __del__(self):
        if self.thisown:
            free_dual_curves(self.len, self)
            

class DualCurve:
    def __init__(self, curves, index):
        self.curves = curves
        self.index = index

    def info(self):
        return my_get_dual_curve_info(self.curves, self.index)

    def __repr__(self):
        return "unfilled %s   filled %s" % self.info()[:2]

    def drill(self, name):
        M = my_drill_cusp(self.curves.manifold, self.curves, self.index, name)
        M.thisown = 1
        return M
    
# Let's try wrapping NormalSurfaceList the right way:

# A NormalSurface really contains no data -- it just "points" to the
# NormalSurfaceListPtr that its a part of.

class NormalSurface:

    def __init__(self, surface_list, index):
        self.surface_list = surface_list
        self.index = index

    def is_orientable(self):
        return normal_surface_is_orientable(self.surface_list, self.index)

    def is_two_sided(self):
        return normal_surface_is_two_sided(self.surface_list, self.index)

    def Euler_characteristic(self):
        return normal_surface_Euler_characteristic(self.surface_list, self.index)

    def __repr__(self):
        euler = self.Euler_characteristic()
        if euler == 2:
            type = "sphere"
        elif euler == 0:
            if self.is_orientable():
                type = "torus"
            else:
                type = "Klein bottle"
        elif euler == 1:
            type = "proj. plane"
        else:
            type = "Euler = %d" % euler
        if self.is_two_sided():
            sided = "Two-sided"
        else:
            sided = "One-sided"
        return sided + " " + type 









